from __future__ import (
    absolute_import,
    division,
    print_function,
)

import json

from pcs import settings
from pcs.common import report_codes
from pcs.lib.commands.sbd import enable_sbd, ALLOWED_SBD_OPTION_LIST
from pcs.test.tools import fixture
from pcs.test.tools.command_env import get_env_tools
from pcs.test.tools.pcs_unittest import TestCase, mock
from pcs.test.tools.misc import get_test_resource, outdent
from pcs.lib.corosync.config_parser import parse_string


def _check_sbd_comm_success_fixture(node, watchdog, device_list):
    return dict(
        label=node,
        output=json.dumps({
            "sbd": {
                "installed": True,
            },
            "watchdog": {
                "exist": True,
                "path": watchdog,
            },
            "device_list": [
                dict(path=dev, exist=True, block_device=True)
                for dev in device_list
            ],
        }),
        param_list=[
            ("watchdog", watchdog),
            ("device_list", json.dumps(device_list)),
        ]
    )


def _get_corosync_conf_text_with_atb(orig_cfg_file):
    corosync_conf = parse_string(open(get_test_resource(orig_cfg_file)).read())
    for quorum in corosync_conf.get_sections(name="quorum"):
        quorum.del_attributes_by_name("two_node")
        quorum.set_attribute("auto_tie_breaker", 1)
    return corosync_conf.export()


def _sbd_enable_successful_report_list_fixture(
    online_node_list, skipped_offline_node_list=(), err_msg="err", atb_set=False
):
    report_list = (
        [
            fixture.warn(report_codes.OMITTING_NODE, node=node)
            for node in skipped_offline_node_list
        ]
        +
        [fixture.info(report_codes.SBD_CHECK_STARTED)]
        +
        [
            fixture.info(report_codes.SBD_CHECK_SUCCESS, node=node)
            for node in online_node_list
        ]
    )
    if atb_set:
        report_list += (
            [
                fixture.warn(report_codes.SBD_REQUIRES_ATB),
                fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
            ]
            +
            [
                fixture.info(
                    report_codes.COROSYNC_NOT_RUNNING_ON_NODE, node=node
                ) for node in online_node_list
            ]
            +
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=node,
                    reason=err_msg,
                    command="remote/status",
                ) for node in skipped_offline_node_list
            ]
            +
            [
                fixture.warn(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                    node=node,
                ) for node in skipped_offline_node_list
            ]
            +
            [fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED)]
            +
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=node,
                    reason=err_msg,
                    command="remote/set_corosync_conf",
                ) for node in skipped_offline_node_list
            ]
            +
            [
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR,
                    node=node,
                ) for node in skipped_offline_node_list
            ]
            +
            [
                fixture.info(
                    report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE, node=node
                ) for node in online_node_list
            ]
        )
    return (
        report_list
        +
        [fixture.info(report_codes.SBD_CONFIG_DISTRIBUTION_STARTED)]
        +
        [
            fixture.info(report_codes.SBD_CONFIG_ACCEPTED_BY_NODE, node=node)
            for node in online_node_list
        ]
        +
        [fixture.info(report_codes.SBD_ENABLING_STARTED)]
        +
        [
            fixture.info(
                report_codes.SERVICE_ENABLE_SUCCESS,
                service="sbd",
                node=node,
                instance=None
            ) for node in online_node_list
        ]
        +
        [fixture.warn(report_codes.CLUSTER_RESTART_REQUIRED_TO_APPLY_CHANGES)]
    )


class OddNumOfNodesSuccess(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.corosync_conf_name = "corosync-3nodes.conf"
        self.node_list = ["rh7-1", "rh7-2", "rh7-3"]
        self.sbd_options = {
            "SBD_WATCHDOG_TIMEOUT": "10",
            "SBD_STARTMODE": "clean",
        }
        self.sbd_config_template = outdent("""\
        # This file has been generated by pcs.
        SBD_DELAY_START=no
        {devices}SBD_OPTS="-n {node_name}"
        SBD_PACEMAKER=yes
        SBD_STARTMODE=clean
        SBD_WATCHDOG_DEV={watchdog}
        SBD_WATCHDOG_TIMEOUT=10
        """)
        self.watchdog_dict = {
            node: "/dev/watchdog-{0}".format(node) for node in self.node_list
        }
        self.config.runner.corosync.version()
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(node_labels=self.node_list)

    def test_with_devices(self):
        device_dict = {
            node: ["/dev/{0}-sbd{1}".format(node, j) for j in range(i)]
            for i, node in enumerate(self.node_list, start=1)
        }
        config_generator = lambda node: self.sbd_config_template.format(
            node_name=node,
            watchdog=self.watchdog_dict[node],
            devices='SBD_DEVICE="{0}"\n'.format(";".join(device_dict[node])),
        )
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(
                    node, self.watchdog_dict[node], device_dict[node]
                ) for node in self.node_list
            ]
        )
        self.config.http.sbd.set_sbd_config(
            config_generator=config_generator, node_labels=self.node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=None,
            watchdog_dict=self.watchdog_dict,
            sbd_options=self.sbd_options,
            default_device_list=[],
            node_device_dict=device_dict,
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(self.node_list)
        )

    def test_no_device(self):
        config_generator = lambda node: self.sbd_config_template.format(
            node_name=node,
            watchdog=self.watchdog_dict[node],
            devices="",
        )
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(
                    node, self.watchdog_dict[node], []
                ) for node in self.node_list
            ]
        )
        self.config.corosync_conf.load(
            filename=self.corosync_conf_name, name="corosync_conf.load2",
        )
        self.config.http.sbd.set_sbd_config(
            config_generator=config_generator, node_labels=self.node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=None,
            watchdog_dict=self.watchdog_dict,
            sbd_options=self.sbd_options,
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(self.node_list)
        )


class OddNumOfNodesDefaultsSuccess(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.corosync_conf_name = "corosync-3nodes.conf"
        self.node_list = ["rh7-1", "rh7-2", "rh7-3"]
        self.sbd_config_template = outdent("""\
        # This file has been generated by pcs.
        SBD_DELAY_START=no
        {devices}SBD_OPTS="-n {node_name}"
        SBD_PACEMAKER=yes
        SBD_STARTMODE=always
        SBD_WATCHDOG_DEV=/dev/watchdog
        SBD_WATCHDOG_TIMEOUT=5
        """)
        self.watchdog = "/dev/watchdog"
        self.config.runner.corosync.version()
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(node_labels=self.node_list)

    def test_with_device(self):
        device_list = ["/dev/sdb"]
        config_generator = lambda node: self.sbd_config_template.format(
            node_name=node,
            devices='SBD_DEVICE="{0}"\n'.format(";".join(device_list)),
        )
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(
                    node, self.watchdog, device_list
                ) for node in self.node_list
            ]
        )
        self.config.http.sbd.set_sbd_config(
            config_generator=config_generator, node_labels=self.node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=self.watchdog,
            watchdog_dict={},
            sbd_options={},
            default_device_list=device_list,
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(self.node_list)
        )

    def test_no_device(self):
        config_generator = lambda node: self.sbd_config_template.format(
            node_name=node, devices="",
        )
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(node, self.watchdog, [])
                for node in self.node_list
            ]
        )
        self.config.corosync_conf.load(
            filename=self.corosync_conf_name, name="corosync_conf.load2",
        )
        self.config.http.sbd.set_sbd_config(
            config_generator=config_generator, node_labels=self.node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=self.watchdog,
            watchdog_dict={},
            sbd_options={},
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(self.node_list)
        )


class WatchdogValidations(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.corosync_conf_name = "corosync-3nodes.conf"
        self.node_list = ["rh7-1", "rh7-2", "rh7-3"]
        self.sbd_config_template = outdent("""\
        # This file has been generated by pcs.
        SBD_DELAY_START=no
        SBD_OPTS="-n {node_name}"
        SBD_PACEMAKER=yes
        SBD_STARTMODE=always
        SBD_WATCHDOG_DEV=/dev/watchdog
        SBD_WATCHDOG_TIMEOUT=5
        """)
        self.watchdog = "/dev/watchdog"
        self.config.runner.corosync.version()
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(node_labels=self.node_list)

    def test_watchdog_not_found(self):
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(
                    node, self.watchdog, []
                ) for node in self.node_list[:2]
            ] + [
                dict(
                    label=self.node_list[2],
                    output=json.dumps({
                        "sbd": {
                            "installed": True,
                        },
                        "watchdog": {
                            "exist": False,
                            "path": self.watchdog,
                        },
                        "device_list": [],
                    }),
                    param_list=[
                        ("watchdog", self.watchdog),
                        ("device_list", "[]"),
                    ]
                )
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            [fixture.info(report_codes.SBD_CHECK_STARTED)]
            +
            [
                fixture.error(
                    report_codes.WATCHDOG_NOT_FOUND,
                    node=self.node_list[2],
                    watchdog=self.watchdog,
                    force_code=report_codes.SKIP_WATCHDOG_VALIDATION,
                )
            ]
            +
            [
                fixture.info(
                    report_codes.SBD_CHECK_SUCCESS, node=node
                ) for node in self.node_list[:2]
            ]
        )

    def test_watchdog_not_supported(self):
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(
                    node, self.watchdog, []
                ) for node in self.node_list[:2]
            ] + [
                dict(
                    label=self.node_list[2],
                    output=json.dumps({
                        "sbd": {
                            "installed": True,
                        },
                        "watchdog": {
                            "exist": True,
                            "path": self.watchdog,
                            "is_supported": False,
                        },
                        "device_list": [],
                    }),
                    param_list=[
                        ("watchdog", self.watchdog),
                        ("device_list", "[]"),
                    ]
                )
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            [fixture.info(report_codes.SBD_CHECK_STARTED)]
            +
            [
                fixture.error(
                    report_codes.SBD_WATCHDOG_NOT_SUPPORTED,
                    node=self.node_list[2],
                    watchdog=self.watchdog,
                    force_code=report_codes.SKIP_WATCHDOG_VALIDATION,
                )
            ]
            +
            [
                fixture.info(
                    report_codes.SBD_CHECK_SUCCESS, node=node
                ) for node in self.node_list[:2]
            ]
        )

    def test_no_watchdog_validation(self):
        config_generator = lambda node: self.sbd_config_template.format(
            node_name=node,
        )
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(node, "", [])
                for node in self.node_list
            ]
        )
        self.config.corosync_conf.load(
            filename=self.corosync_conf_name, name="corosync_conf.load2",
        )
        self.config.http.sbd.set_sbd_config(
            config_generator=config_generator, node_labels=self.node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=self.watchdog,
            watchdog_dict={},
            sbd_options={},
            no_watchdog_validation=True,
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(self.node_list) + [
                fixture.warn(report_codes.SBD_WATCHDOG_VALIDATION_INACTIVE)
            ]
        )


class EvenNumOfNodes(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.corosync_conf_name = "corosync.conf"
        self.node_list = ["rh7-1", "rh7-2"]
        self.sbd_config_template = outdent("""\
        # This file has been generated by pcs.
        SBD_DELAY_START=no
        {devices}SBD_OPTS="-n {node_name}"
        SBD_PACEMAKER=yes
        SBD_STARTMODE=always
        SBD_WATCHDOG_DEV=/dev/watchdog
        SBD_WATCHDOG_TIMEOUT=5
        """)
        self.watchdog = "/dev/watchdog"
        self.config.runner.corosync.version()
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(node_labels=self.node_list)

    def test_with_device(self):
        device_list = ["/dev/sdb"]
        config_generator = lambda node: self.sbd_config_template.format(
            node_name=node,
            devices='SBD_DEVICE="{0}"\n'.format(";".join(device_list)),
        )
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(
                    node, self.watchdog, device_list
                ) for node in self.node_list
            ]
        )
        self.config.http.sbd.set_sbd_config(
            config_generator=config_generator, node_labels=self.node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=self.watchdog,
            watchdog_dict={},
            sbd_options={},
            default_device_list=device_list,
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(self.node_list)
        )

    @mock.patch("pcs.lib.external.is_systemctl", lambda: True)
    def test_no_device(self):
        config_generator = lambda node: self.sbd_config_template.format(
            node_name=node, devices="",
        )
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(node, self.watchdog, [])
                for node in self.node_list
            ]
        )
        self.config.corosync_conf.load(
            filename=self.corosync_conf_name, name="corosync_conf.load2",
        )
        self.config.http.corosync.check_corosync_offline(
            node_labels=self.node_list
        )
        self.config.http.corosync.set_corosync_conf(
            _get_corosync_conf_text_with_atb(self.corosync_conf_name),
            node_labels=self.node_list,
        )
        self.config.runner.systemctl.is_active("corosync", is_active=False)
        self.config.http.sbd.set_sbd_config(
            config_generator=config_generator, node_labels=self.node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=self.watchdog,
            watchdog_dict={},
            sbd_options={},
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True,
            )
        )

    def test_no_device_auto_tie_breaker_enabled(self):
        config_generator = lambda node: self.sbd_config_template.format(
            node_name=node, devices="",
        )
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(node, self.watchdog, [])
                for node in self.node_list
            ]
        )
        self.config.corosync_conf.load(
            filename=self.corosync_conf_name, name="corosync_conf.load2",
            auto_tie_breaker=True,
        )
        self.config.http.sbd.set_sbd_config(
            config_generator=config_generator, node_labels=self.node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=self.watchdog,
            watchdog_dict={},
            sbd_options={},
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(self.node_list)
        )

    def test_no_device_with_qdevice(self):
        config_generator = lambda node: self.sbd_config_template.format(
            node_name=node, devices="",
        )
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(node, self.watchdog, [])
                for node in self.node_list
            ]
        )
        self.config.corosync_conf.load(
            filename="corosync-qdevice.conf", name="corosync_conf.load2",
        )
        self.config.http.sbd.set_sbd_config(
            config_generator=config_generator, node_labels=self.node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=self.watchdog,
            watchdog_dict={},
            sbd_options={},
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(self.node_list)
        )

class OfflineNodes(TestCase):
    #pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.corosync_conf_name = "corosync.conf"
        node_list = ["rh7-1", "rh7-2"]
        self.online_node_list = node_list[:-1]
        self.offline_node_list = node_list[-1:]
        self.watchdog = "/dev/watchdog"
        self.err_msg = "error msg"
        self.sbd_config_generator = outdent("""\
        # This file has been generated by pcs.
        SBD_DELAY_START=no
        SBD_OPTS="-n {0}"
        SBD_PACEMAKER=yes
        SBD_STARTMODE=always
        SBD_WATCHDOG_DEV=/dev/watchdog
        SBD_WATCHDOG_TIMEOUT=5
        """).format
        self.offline_communication_list = (
            [dict(label=node) for node in self.online_node_list]
            +
            [
                dict(
                    label=node,
                    was_connected=False,
                    errno=1,
                    error_msg=self.err_msg,
                ) for node in self.offline_node_list
            ]
        )
        self.config.runner.corosync.version()
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(
            communication_list=self.offline_communication_list
        )

    def test_no_ignore_offline_nodes(self):
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=None,
                watchdog_dict={},
                sbd_options={},
            ),
            [],
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    node=node,
                    command="remote/check_auth",
                    reason=self.err_msg,
                ) for node in self.offline_node_list
            ]
        )

    def test_ignore_offline_nodes(self):
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(node, self.watchdog, [])
                for node in self.online_node_list
            ]
        )
        self.config.corosync_conf.load(
            filename="corosync-qdevice.conf", name="corosync_conf.load2",
        )
        self.config.http.sbd.set_sbd_config(
            config_generator=self.sbd_config_generator,
            node_labels=self.online_node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.online_node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.online_node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=None,
            watchdog_dict={},
            sbd_options={},
            ignore_offline_nodes=True,
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.OMITTING_NODE,
                    node=node,
                ) for node in self.offline_node_list
            ]
            +
            _sbd_enable_successful_report_list_fixture(self.online_node_list)
        )

    @mock.patch("pcs.lib.external.is_systemctl", lambda: True)
    def test_ignore_offline_nodes_atb_needed(self):
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(node, self.watchdog, [])
                for node in self.online_node_list
            ]
        )
        self.config.corosync_conf.load(
            filename=self.corosync_conf_name, name="corosync_conf.load2",
        )
        self.config.http.corosync.check_corosync_offline(
            communication_list=self.offline_communication_list
        )
        self.config.http.corosync.set_corosync_conf(
            _get_corosync_conf_text_with_atb(self.corosync_conf_name),
            communication_list=self.offline_communication_list,
        )
        self.config.runner.systemctl.is_active("corosync", is_active=False)
        self.config.http.sbd.set_sbd_config(
            config_generator=self.sbd_config_generator,
            node_labels=self.online_node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.online_node_list[0]]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.online_node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=None,
            watchdog_dict={},
            sbd_options={},
            ignore_offline_nodes=True,
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.online_node_list,
                skipped_offline_node_list=self.offline_node_list,
                err_msg=self.err_msg,
                atb_set=True,
            )
        )

# watchdog validations are in WatchdogValidations
class Validations(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.corosync_conf_name = "corosync.conf"
        self.node_list = ["rh7-1", "rh7-2"]
        self.config.runner.corosync.version()
        self.config.corosync_conf.load(filename=self.corosync_conf_name)

    def test_non_existing_node_in_watchdogs(self):
        unknown_node = "unknown_node"
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=None,
                watchdog_dict={
                    node: "/dev/watchdog"
                    for node in (self.node_list + [unknown_node])
                },
                sbd_options={},
            ),
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    node=unknown_node,
                    searched_types=[],
                )
            ],
        )
        self.env_assist.assert_reports([])

    def test_non_existing_node_in_devices(self):
        unknown_node = "unknown_node"
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog="/dev/watchdog",
                watchdog_dict={},
                sbd_options={},
                default_device_list="/device",
                node_device_dict={
                    node: ["/device"]
                    for node in (self.node_list + [unknown_node])
                }
            ),
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    node=unknown_node,
                    searched_types=[],
                )
            ],
        )
        self.env_assist.assert_reports([])

    def test_watchdog_not_abs_path(self):
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog="wd1",
                watchdog_dict={self.node_list[0]: "wd2"},
                sbd_options={},
            ),
            [
                fixture.error(
                    report_codes.WATCHDOG_INVALID,
                    watchdog=w,
                ) for w in ["wd1", "wd2"]
            ],
        )
        self.env_assist.assert_reports([])

    def test_device_not_abs_path(self):
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog="/dev/watchdog",
                watchdog_dict={},
                sbd_options={},
                default_device_list=["device1"],
                node_device_dict={self.node_list[0]: ["device2"]}
            ),
            [
                fixture.error(
                    report_codes.SBD_DEVICE_PATH_NOT_ABSOLUTE,
                    node=node,
                    device=dev,
                ) for node, dev in [
                    (self.node_list[0], "device2"),
                    (self.node_list[1], "device1"),
                ]
            ],
        )
        self.env_assist.assert_reports([])

    def test_no_device_for_node(self):
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog="/dev/watchdog",
                watchdog_dict={},
                sbd_options={},
                default_device_list=[],
                node_device_dict={self.node_list[0]: ["/dev/device1"]}
            ),
            [
                fixture.error(
                    report_codes.SBD_NO_DEVICE_FOR_NODE,
                    node=self.node_list[1],
                )
            ],
        )
        self.env_assist.assert_reports([])

    def test_too_many_devices(self):
        max_dev_num = settings.sbd_max_device_num
        dev_list = ["/dev/dev{0}".format(i) for i in range(max_dev_num + 1)]
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog="/dev/watchdog",
                watchdog_dict={},
                sbd_options={},
                default_device_list=["/dev/dev1"],
                node_device_dict={self.node_list[0]: dev_list}
            ),
            [
                fixture.error(
                    report_codes.SBD_TOO_MANY_DEVICES_FOR_NODE,
                    node=self.node_list[0],
                    device_list=dev_list,
                    max_devices=max_dev_num,
                )
            ],
        )
        self.env_assist.assert_reports([])

    def test_unknown_sbd_opts(self):
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog="/dev/watchdog",
                watchdog_dict={},
                sbd_options={
                    "UNKNOWN_OPT1": 1,
                    "SBD_STARTMODE": "clean",
                    "UNKNOWN_OPT2": "val",
                    "SBD_WATCHDOG_DEV": "dev",
                },
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=[opt],
                    option_type=None,
                    allowed=sorted(ALLOWED_SBD_OPTION_LIST),
                    allowed_patterns=[],
                    force_code=report_codes.FORCE_OPTIONS,
                ) for opt in ["UNKNOWN_OPT1", "UNKNOWN_OPT2"]
            ]
            +
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["SBD_WATCHDOG_DEV"],
                    option_type=None,
                    allowed=sorted(ALLOWED_SBD_OPTION_LIST),
                    allowed_patterns=[],
                ),
            ]
        )
        self.env_assist.assert_reports([])

    def test_unknown_sbd_opts_allowed(self):
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog="/dev/watchdog",
                watchdog_dict={},
                sbd_options={
                    "UNKNOWN_OPT1": 1,
                    "SBD_STARTMODE": "clean",
                    "UNKNOWN_OPT2": "val",
                    "SBD_WATCHDOG_DEV": "dev",
                },
                allow_unknown_opts=True,
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["SBD_WATCHDOG_DEV"],
                    option_type=None,
                    allowed=sorted(ALLOWED_SBD_OPTION_LIST),
                    allowed_patterns=[],
                ),
            ]
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.INVALID_OPTIONS,
                    option_names=[opt],
                    option_type=None,
                    allowed=sorted(ALLOWED_SBD_OPTION_LIST),
                    allowed_patterns=[],
                ) for opt in ["UNKNOWN_OPT1", "UNKNOWN_OPT2"]
            ]
        )

    def test_sbd_not_installed(self):
        watchdog = "/dev/watchdog"
        self.config.http.host.check_auth(node_labels=self.node_list)
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(
                    self.node_list[0], watchdog, []
                ),
                dict(
                    label=self.node_list[1],
                    output=json.dumps({
                        "sbd": {
                            "installed": False,
                        },
                        "watchdog": {
                            "exist": True,
                            "path": watchdog,
                        },
                        "device_list": [],
                    }),
                    param_list=[
                        ("watchdog", watchdog),
                        ("device_list", "[]"),
                    ]
                )
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            [fixture.info(report_codes.SBD_CHECK_STARTED)]
            +
            [
                fixture.error(
                    report_codes.SBD_NOT_INSTALLED, node=self.node_list[1]
                )
            ]
            +
            [
                fixture.info(
                    report_codes.SBD_CHECK_SUCCESS, node=self.node_list[0]
                )
            ]
        )

    def test_device_not_exists_not_block_device(self):
        watchdog = "/dev/watchdog"
        device_list = ["/dev/dev0", "/dev/dev1"]
        self.config.http.host.check_auth(node_labels=self.node_list)
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(
                    self.node_list[0], watchdog, device_list
                ),
                dict(
                    label=self.node_list[1],
                    output=json.dumps({
                        "sbd": {
                            "installed": True,
                        },
                        "watchdog": {
                            "exist": True,
                            "path": watchdog,
                        },
                        "device_list": [
                            dict(
                                path=device_list[0],
                                exist=False,
                                block_device=False,
                            ),
                            dict(
                                path=device_list[1],
                                exist=True,
                                block_device=False,
                            ),
                        ],
                    }),
                    param_list=[
                        ("watchdog", watchdog),
                        ("device_list", json.dumps(device_list)),
                    ]
                )
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=watchdog,
                watchdog_dict={},
                sbd_options={},
                default_device_list=device_list,
            ),
            []
        )
        self.env_assist.assert_reports(
            [fixture.info(report_codes.SBD_CHECK_STARTED)]
            +
            [
                fixture.error(
                    report_codes.SBD_DEVICE_DOES_NOT_EXIST,
                    node=self.node_list[1],
                    device=device_list[0],
                ),
                fixture.error(
                    report_codes.SBD_DEVICE_IS_NOT_BLOCK_DEVICE,
                    node=self.node_list[1],
                    device=device_list[1],
                )
            ]
            +
            [
                fixture.info(
                    report_codes.SBD_CHECK_SUCCESS, node=self.node_list[0]
                )
            ]
        )

    def test_multiple_validation_failures(self):
        unknown_node_list = ["unknown_node{0}".format(i) for i in range(2)]
        max_dev_num = settings.sbd_max_device_num
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog="watchdog0",
                watchdog_dict={
                    unknown_node_list[0]: "/dev/watchdog",
                },
                sbd_options={
                    "UNKNOWN_OPT1": 1,
                    "SBD_STARTMODE": "clean",
                    "UNKNOWN_OPT2": "val",
                    "SBD_WATCHDOG_DEV": "dev",
                },
                default_device_list=[],
                node_device_dict={
                    self.node_list[0]: ["dev", "/dev0", "/dev1", "/dev2"],
                    unknown_node_list[0]: ["/dev/device0"],
                    unknown_node_list[1]: ["/dev/device0"],
                }
            ),
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    node=node,
                    searched_types=[],
                ) for node in unknown_node_list
            ]
            +
            [
                fixture.error(
                    report_codes.WATCHDOG_INVALID, watchdog="watchdog0"
                ),
                fixture.error(
                    report_codes.WATCHDOG_INVALID, watchdog="watchdog0"
                ),
                fixture.error(
                    report_codes.SBD_NO_DEVICE_FOR_NODE, node=self.node_list[1],
                ),
                fixture.error(
                    report_codes.SBD_TOO_MANY_DEVICES_FOR_NODE,
                    node=self.node_list[0],
                    device_list=["dev", "/dev0", "/dev1", "/dev2"],
                    max_devices=max_dev_num,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["SBD_WATCHDOG_DEV"],
                    option_type=None,
                    allowed=sorted(ALLOWED_SBD_OPTION_LIST),
                    allowed_patterns=[],
                ),
            ]
            +
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=[opt],
                    option_type=None,
                    allowed=sorted(ALLOWED_SBD_OPTION_LIST),
                    allowed_patterns=[],
                    force_code=report_codes.FORCE_OPTIONS,
                ) for opt in ["UNKNOWN_OPT1", "UNKNOWN_OPT2"]
            ]
        )
        self.env_assist.assert_reports([])


class FailureHandling(TestCase):
    #pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.corosync_conf_name = "corosync.conf"
        self.node_list = ["rh7-1", "rh7-2"]
        self.sbd_config_generator = outdent("""\
        # This file has been generated by pcs.
        SBD_DELAY_START=no
        SBD_OPTS="-n {0}"
        SBD_PACEMAKER=yes
        SBD_STARTMODE=always
        SBD_WATCHDOG_DEV=/dev/watchdog
        SBD_WATCHDOG_TIMEOUT=5
        """).format
        self.watchdog = "/dev/watchdog"
        self.reason = "failure reason"
        self.communication_list_failure = [
            dict(
                label=self.node_list[0],
                response_code=400,
                output=self.reason,
            ),
            dict(
                label=self.node_list[1],
            )
        ]
        self.communication_list_not_connected = [
            dict(
                label=self.node_list[0],
                errno=1,
                error_msg=self.reason,
                was_connected=False,
            ),
            dict(
                label=self.node_list[1],
            )
        ]
        self.config.runner.corosync.version()
        self.config.corosync_conf.load(filename=self.corosync_conf_name)
        self.config.http.host.check_auth(node_labels=self.node_list)
        self.config.http.sbd.check_sbd(
            communication_list=[
                _check_sbd_comm_success_fixture(node, self.watchdog, [])
                for node in self.node_list
            ]
        )
        self.config.corosync_conf.load(
            filename=self.corosync_conf_name, name="corosync_conf.load2",
        )
        self.config.http.corosync.check_corosync_offline(
            node_labels=self.node_list
        )
        self.config.http.corosync.set_corosync_conf(
            _get_corosync_conf_text_with_atb(self.corosync_conf_name),
            node_labels=self.node_list,
        )
        self.config.runner.systemctl.is_active("corosync", is_active=False)
        self.config.http.sbd.set_sbd_config(
            config_generator=self.sbd_config_generator,
            node_labels=self.node_list,
        )
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            node_labels=[self.node_list[0]]
        )

    def _remove_calls(self, n):
        for name in self.config.calls.names[-n:]:
            self.config.calls.remove(name)

    @mock.patch("pcs.lib.external.is_systemctl", lambda: True)
    def test_enable_failed(self):
        self.config.http.sbd.enable_sbd(
            communication_list=self.communication_list_failure
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True
            )[:-3]
            +
            [
                fixture.info(
                    report_codes.SERVICE_ENABLE_SUCCESS,
                    service="sbd",
                    node=self.node_list[1],
                    instance=None
                ),
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/sbd_enable"
                )
            ]
        )

    @mock.patch("pcs.lib.external.is_systemctl", lambda: True)
    def test_enable_not_connected(self):
        self.config.http.sbd.enable_sbd(
            communication_list=self.communication_list_not_connected
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True
            )[:-3]
            +
            [
                fixture.info(
                    report_codes.SERVICE_ENABLE_SUCCESS,
                    service="sbd",
                    node=self.node_list[1],
                    instance=None
                ),
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/sbd_enable"
                )
            ]
        )

    @mock.patch("pcs.lib.external.is_systemctl", lambda: True)
    def test_removing_stonith_wd_timeout_failure(self):
        self._remove_calls(2)
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            communication_list=[
                self.communication_list_failure[:1],
                [dict(label=self.node_list[1])]
            ]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=self.watchdog,
            watchdog_dict={},
            sbd_options={},
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True
            )
            +
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/remove_stonith_watchdog_timeout",
                )
            ]
        )

    @mock.patch("pcs.lib.external.is_systemctl", lambda: True)
    def test_removing_stonith_wd_timeout_not_connected(self):
        self._remove_calls(2)
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            communication_list=[
                self.communication_list_not_connected[:1],
                [dict(label=self.node_list[1])]
            ]
        )
        self.config.http.sbd.enable_sbd(node_labels=self.node_list)
        enable_sbd(
            self.env_assist.get_env(),
            default_watchdog=self.watchdog,
            watchdog_dict={},
            sbd_options={},
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True
            )
            +
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/remove_stonith_watchdog_timeout",
                )
            ]
        )

    @mock.patch("pcs.lib.external.is_systemctl", lambda: True)
    def test_removing_stonith_wd_timeout_complete_failure(self):
        self._remove_calls(2)
        self.config.http.pcmk.remove_stonith_watchdog_timeout(
            communication_list=[
                self.communication_list_not_connected[:1],
                [dict(
                    label=self.node_list[1],
                    response_code=400,
                    output=self.reason,
                )],
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True
            )[:-4]
            +
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/remove_stonith_watchdog_timeout",
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_list[1],
                    reason=self.reason,
                    command="remote/remove_stonith_watchdog_timeout",
                ),
                fixture.error(
                    report_codes.UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE,
                )
            ]
        )

    @mock.patch("pcs.lib.external.is_systemctl", lambda: True)
    def test_set_sbd_config_failure(self):
        self._remove_calls(4)
        self.config.http.sbd.set_sbd_config(
            communication_list=[
                dict(
                    label=self.node_list[0],
                    param_list=[
                        ("config", self.sbd_config_generator(self.node_list[0]))
                    ],
                    response_code=400,
                    output=self.reason,
                ),
                dict(
                    label=self.node_list[1],
                    param_list=[
                        ("config", self.sbd_config_generator(self.node_list[1]))
                    ],
                ),
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True
            )[:-6]
            +
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/set_sbd_config",
                ),
                fixture.info(
                    report_codes.SBD_CONFIG_ACCEPTED_BY_NODE,
                    node=self.node_list[1],
                )
            ]
        )

    def test_set_corosync_conf_failed(self):
        self._remove_calls(7)
        self.config.http.corosync.set_corosync_conf(
            _get_corosync_conf_text_with_atb(self.corosync_conf_name),
            communication_list=self.communication_list_failure,
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True
            )[:-9]
            +
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/set_corosync_conf",
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.error(
                    report_codes.COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR,
                    node=self.node_list[0],
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.info(
                    report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                    node=self.node_list[1],
                )
            ]
        )

    def test_set_corosync_conf_not_connected(self):
        self._remove_calls(7)
        self.config.http.corosync.set_corosync_conf(
            _get_corosync_conf_text_with_atb(self.corosync_conf_name),
            communication_list=self.communication_list_not_connected,
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True
            )[:-9]
            +
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/set_corosync_conf",
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.error(
                    report_codes.COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR,
                    node=self.node_list[0],
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.info(
                    report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                    node=self.node_list[1],
                )
            ]
        )

    def test_corosync_not_running_failed(self):
        self._remove_calls(9)
        self.config.http.corosync.check_corosync_offline(
            communication_list=self.communication_list_failure,
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True
            )[:-12]
            +
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/status",
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.error(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                    node=self.node_list[0],
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.info(
                    report_codes.COROSYNC_NOT_RUNNING_ON_NODE,
                    node=self.node_list[1]
                )
            ]
        )

    def test_corosync_not_running_not_connected(self):
        self._remove_calls(9)
        self.config.http.corosync.check_corosync_offline(
            communication_list=self.communication_list_not_connected,
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            _sbd_enable_successful_report_list_fixture(
                self.node_list, atb_set=True
            )[:-12]
            +
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/status",
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.error(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                    node=self.node_list[0],
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                ),
                fixture.info(
                    report_codes.COROSYNC_NOT_RUNNING_ON_NODE,
                    node=self.node_list[1]
                )
            ]
        )

    def test_check_sbd_invalid_data_format(self):
        self._remove_calls(12)
        self.config.http.sbd.check_sbd(
            communication_list=[
                dict(
                    label=self.node_list[0],
                    param_list=[
                        ("watchdog", self.watchdog),
                        ("device_list", "[]"),
                    ],
                    output="{}",
                ),
                dict(
                    label=self.node_list[1],
                    param_list=[
                        ("watchdog", self.watchdog),
                        ("device_list", "[]"),
                    ],
                    output="not JSON",
                ),
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            [fixture.info(report_codes.SBD_CHECK_STARTED)]
            +
            [
                fixture.error(report_codes.INVALID_RESPONSE_FORMAT, node=node)
                for node in self.node_list
            ]
        )

    def test_check_sbd_failure(self):
        self._remove_calls(12)
        self.config.http.sbd.check_sbd(
            communication_list=[
                dict(
                    label=self.node_list[0],
                    param_list=[
                        ("watchdog", self.watchdog),
                        ("device_list", "[]"),
                    ],
                    response_code=400,
                    output=self.reason,
                ),
                _check_sbd_comm_success_fixture(
                    self.node_list[1], self.watchdog, []
                )
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.info(report_codes.SBD_CHECK_STARTED),
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/check_sbd",
                ),
                fixture.info(
                    report_codes.SBD_CHECK_SUCCESS,
                    node=self.node_list[1]
                )
            ]
        )

    def test_check_sbd_not_connected(self):
        self._remove_calls(12)
        self.config.http.sbd.check_sbd(
            communication_list=[
                dict(
                    label=self.node_list[0],
                    param_list=[
                        ("watchdog", self.watchdog),
                        ("device_list", "[]"),
                    ],
                    errno=1,
                    error_msg=self.reason,
                    was_connected=False,
                ),
                _check_sbd_comm_success_fixture(
                    self.node_list[1], self.watchdog, []
                )
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.info(report_codes.SBD_CHECK_STARTED),
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/check_sbd",
                ),
                fixture.info(
                    report_codes.SBD_CHECK_SUCCESS,
                    node=self.node_list[1]
                )
            ]
        )

    def test_get_online_targets_failed(self):
        self._remove_calls(14)
        self.config.http.host.check_auth(
            communication_list=self.communication_list_failure
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/check_auth",
                )
            ]
        )

    def test_get_online_targets_not_connected(self):
        self._remove_calls(14)
        self.config.http.host.check_auth(
            communication_list=self.communication_list_not_connected
        )
        self.env_assist.assert_raise_library_error(
            lambda: enable_sbd(
                self.env_assist.get_env(),
                default_watchdog=self.watchdog,
                watchdog_dict={},
                sbd_options={},
            ),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.node_list[0],
                    reason=self.reason,
                    command="remote/check_auth",
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                )
            ]
        )



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from functools import partial
from lxml import etree

from pcs.common import report_codes
from pcs.lib.cib.resource import operations
from pcs.lib.errors import ReportItemSeverity as severities
from pcs.lib.validate import ValuePair
from pcs.test.tools.assertions import assert_report_item_list_equal
from pcs.test.tools.custom_mock import MockLibraryReportProcessor
from pcs.test.tools.misc import create_patcher
from pcs.test.tools.pcs_unittest import TestCase, mock


patch_operations = create_patcher("pcs.lib.cib.resource.operations")

@patch_operations("get_remaining_defaults")
@patch_operations("complete_all_intervals")
@patch_operations("validate_different_intervals")
@patch_operations("validate_operation_list")
@patch_operations("normalized_to_operations")
@patch_operations("operations_to_normalized")
class Prepare(TestCase):
    def test_prepare(
        self, operations_to_normalized, normalized_to_operations,
        validate_operation_list, validate_different_intervals,
        complete_all_intervals, get_remaining_defaults
    ):
        validate_operation_list.return_value = ["options_report"]
        validate_different_intervals.return_value = [
            "different_interval_report"
        ]
        operations_to_normalized.return_value = [
            {"name": ValuePair("Start", "start")},
            {"name": ValuePair("Monitor", "monitor")},
        ]
        normalized_to_operations.return_value = [
            {"name": "start"},
            {"name": "monitor"},
        ]

        report_processor = mock.MagicMock()
        raw_operation_list = [
            {"name": "Start"},
            {"name": "Monitor"},
        ]
        default_operation_list = [
            {"name": "stop"},
        ]
        allowed_operation_name_list = ["start", "stop", "monitor"]
        allow_invalid = True

        operations.prepare(
            report_processor,
            raw_operation_list,
            default_operation_list,
            allowed_operation_name_list,
            allow_invalid,
        )

        operations_to_normalized.assert_called_once_with(raw_operation_list)
        normalized_to_operations.assert_called_once_with(
            operations_to_normalized.return_value
        )
        validate_operation_list.assert_called_once_with(
            operations_to_normalized.return_value,
            allowed_operation_name_list,
            allow_invalid
        )
        validate_different_intervals.assert_called_once_with(
            normalized_to_operations.return_value
        )
        complete_all_intervals.assert_called_once_with(
            normalized_to_operations.return_value
        )
        get_remaining_defaults.assert_called_once_with(
            report_processor,
            normalized_to_operations.return_value,
            default_operation_list
        )
        report_processor.process_list.assert_called_once_with([
            "options_report",
            "different_interval_report",
        ])


class ValidateDifferentIntervals(TestCase):
    def test_return_empty_reports_on_empty_list(self):
        operations.validate_different_intervals([])

    def test_return_empty_reports_on_operations_without_duplication(self):
        operations.validate_different_intervals([
            {"name": "monitor", "interval": "10s"},
            {"name": "monitor", "interval": "5s"},
            {"name": "start", "interval": "5s"},
        ])

    def test_return_report_on_duplicated_intervals(self):
        assert_report_item_list_equal(
            operations.validate_different_intervals([
                {"name": "monitor", "interval": "3600s"},
                {"name": "monitor", "interval": "60m"},
                {"name": "monitor", "interval": "1h"},
                {"name": "monitor", "interval": "60s"},
                {"name": "monitor", "interval": "1m"},
                {"name": "monitor", "interval": "5s"},
            ]),
            [(
                severities.ERROR,
                report_codes.RESOURCE_OPERATION_INTERVAL_DUPLICATION,
                {
                    "duplications": {
                        "monitor": [
                            ["3600s", "60m", "1h"],
                            ["60s", "1m"],
                        ],
                    },
                },
            )]
        )

class MakeUniqueIntervals(TestCase):
    def setUp(self):
        self.report_processor = MockLibraryReportProcessor()
        self.run = partial(
            operations.make_unique_intervals,
            self.report_processor
        )

    def test_return_copy_input_when_no_interval_duplication(self):
        operation_list = [
            {"name": "monitor", "interval": "10s"},
            {"name": "monitor", "interval": "5s"},
            {"name": "monitor", },
            {"name": "monitor", "interval": ""},
            {"name": "start", "interval": "5s"},
        ]
        self.assertEqual(operation_list, self.run(operation_list))

    def test_adopt_duplicit_values(self):
        self.assertEqual(
            self.run([
                {"name": "monitor", "interval": "60s"},
                {"name": "monitor", "interval": "1m"},
                {"name": "monitor", "interval": "5s"},
                {"name": "monitor", "interval": "6s"},
                {"name": "monitor", "interval": "5s"},
                {"name": "start", "interval": "5s"},
            ]),
            [
                {"name": "monitor", "interval": "60s"},
                {"name": "monitor", "interval": "61"},
                {"name": "monitor", "interval": "5s"},
                {"name": "monitor", "interval": "6s"},
                {"name": "monitor", "interval": "7"},
                {"name": "start", "interval": "5s"},
            ]
        )

        assert_report_item_list_equal(self.report_processor.report_item_list, [
            (
                severities.WARNING,
                report_codes.RESOURCE_OPERATION_INTERVAL_ADAPTED,
                {
                    "operation_name": "monitor",
                    "original_interval": "1m",
                    "adapted_interval": "61",
                },
            ),
            (
                severities.WARNING,
                report_codes.RESOURCE_OPERATION_INTERVAL_ADAPTED,
                {
                    "operation_name": "monitor",
                    "original_interval": "5s",
                    "adapted_interval": "7",
                },
            ),
        ])

    def test_keep_duplicit_values_when_are_not_valid_interval(self):
        self.assertEqual(
            self.run([
                {"name": "monitor", "interval": "some"},
                {"name": "monitor", "interval": "some"},
            ]),
            [
                {"name": "monitor", "interval": "some"},
                {"name": "monitor", "interval": "some"},
            ]
        )


class Normalize(TestCase):
    def test_return_operation_with_the_same_values(self):
        operation = {
            "name": "monitor",
            "role": "Master",
            "timeout": "10",
        }

        self.assertEqual(operation, dict([
            (key, operations.normalize(key, value))
            for key, value in operation.items()
        ]))

    def test_return_operation_with_normalized_values(self):
        self.assertEqual(
            {
                "name": "monitor",
                "role": "Master",
                "timeout": "10",
                "requires": "nothing",
                "on-fail": "ignore",
                "record-pending": "true",
                "enabled": "1",
            },
            dict([(key, operations.normalize(key, value)) for key, value in {
                "name": "monitor",
                "role": "master",
                "timeout": "10",
                "requires": "Nothing",
                "on-fail": "Ignore",
                "record-pending": "True",
                "enabled": "1",
            }.items()])
        )

class ValidateOperation(TestCase):
    def assert_operation_produces_report(self, operation, report_list):
        assert_report_item_list_equal(
            operations.validate_operation_list(
                [operation],
                ["monitor"],
            ),
            report_list
        )

    def test_return_empty_report_on_valid_operation(self):
        self.assert_operation_produces_report(
            {
                "name": "monitor",
                "role": "Master"
            },
            []
        )

    def test_validate_all_individual_options(self):
        self.assertEqual(
            ["REQUIRES REPORT", "ROLE REPORT"],
            sorted(operations.validate_operation({"name": "monitor"}, [
                mock.Mock(return_value=["ROLE REPORT"]),
                mock.Mock(return_value=["REQUIRES REPORT"]),
            ]))
        )

    def test_return_error_when_unknown_operation_attribute(self):
        self.assert_operation_produces_report(
            {
                "name": "monitor",
                "unknown": "invalid",
            },
            [
                (
                    severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["unknown"],
                        "option_type": "resource operation",
                        "allowed": sorted(operations.ATTRIBUTES),
                        "allowed_patterns": [],
                    },
                    None
                ),
            ],
        )

    def test_return_errror_when_missing_key_name(self):
        self.assert_operation_produces_report(
            {
                "role": "Master"
            },
            [
                (
                    severities.ERROR,
                    report_codes.REQUIRED_OPTION_IS_MISSING,
                    {
                        "option_names": ["name"],
                        "option_type": "resource operation",
                    },
                    None
                ),
            ],
        )

    def test_return_error_when_both_interval_origin_and_start_delay(self):
        self.assert_operation_produces_report(
            {
                "name": "monitor",
                "interval-origin": "a",
                "start-delay": "b",
            },
            [
                (
                    severities.ERROR,
                    report_codes.MUTUALLY_EXCLUSIVE_OPTIONS,
                    {
                        "option_names": ["interval-origin", "start-delay"],
                        "option_type": "resource operation",
                    },
                    None
                ),
            ],
        )

    def test_return_error_on_invalid_id(self):
        self.assert_operation_produces_report(
            {
                "name": "monitor",
                "id": "a#b",
            },
            [
                (
                    severities.ERROR,
                    report_codes.INVALID_ID,
                    {
                        "id": "a#b",
                        "id_description": "operation id",
                        "invalid_character": "#",
                        "is_first_char": False,
                    },
                    None
                ),
            ],
        )


class GetRemainingDefaults(TestCase):
    @mock.patch("pcs.lib.cib.resource.operations.make_unique_intervals")
    def test_returns_remining_operations(self, make_unique_intervals):
        make_unique_intervals.side_effect = (
            lambda report_processor, operations: operations
        )
        self.assertEqual(
            operations.get_remaining_defaults(
                report_processor=None,
                operation_list =[{"name": "monitor"}],
                default_operation_list=[{"name": "monitor"}, {"name": "start"}]
            ),
            [{"name": "start"}]
        )


class GetResourceOperations(TestCase):
    resource_el = etree.fromstring("""
        <primitive class="ocf" id="dummy" provider="pacemaker" type="Stateful">
            <operations>
                <op id="dummy-start" interval="0s" name="start" timeout="20"/>
                <op id="dummy-stop" interval="0s" name="stop" timeout="20"/>
                <op id="dummy-monitor-m" interval="10" name="monitor"
                    role="Master" timeout="20"/>
                <op id="dummy-monitor-s" interval="11" name="monitor"
                    role="Slave" timeout="20"/>
            </operations>
        </primitive>
    """)
    resource_noop_el = etree.fromstring("""
        <primitive class="ocf" id="dummy" provider="pacemaker" type="Stateful">
        </primitive>
    """)

    def assert_op_list(self, op_list, expected_ids):
        self.assertEqual(
            [op.attrib.get("id") for op in op_list],
            expected_ids
        )

    def test_all_operations(self):
        self.assert_op_list(
            operations.get_resource_operations(self.resource_el),
            ["dummy-start", "dummy-stop", "dummy-monitor-m", "dummy-monitor-s"]
        )

    def test_filter_operations(self):
        self.assert_op_list(
            operations.get_resource_operations(self.resource_el, ["start"]),
            ["dummy-start"]
        )

    def test_filter_more_operations(self):
        self.assert_op_list(
            operations.get_resource_operations(
                self.resource_el,
                ["monitor", "stop"]
            ),
            ["dummy-stop", "dummy-monitor-m", "dummy-monitor-s"]
        )

    def test_filter_none(self):
        self.assert_op_list(
            operations.get_resource_operations(self.resource_el, ["promote"]),
            []
        )

    def test_no_operations(self):
        self.assert_op_list(
            operations.get_resource_operations(self.resource_noop_el),
            []
        )



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from pcs.test.tools.pcs_unittest import TestCase

from pcs.cli.booth import env
from pcs.common import report_codes, env_file_role_codes
from pcs.lib.errors import LibraryEnvError, ReportItem
from pcs.test.tools.pcs_unittest import mock
from pcs.test.tools.misc import create_setup_patch_mixin

SetupPatchMixin = create_setup_patch_mixin(env)

class BoothConfTest(TestCase, SetupPatchMixin):
    def setUp(self):
        self.write = self.setup_patch("env_file.write")
        self.read = self.setup_patch("env_file.read")
        self.process_no_existing_file_expectation = self.setup_patch(
            "env_file.process_no_existing_file_expectation"
        )

    def test_sucessfully_care_about_local_file(self):
        def next_in_line(env):
            env.booth["modified_env"] = {
                "config_file": {
                    "content": "file content",
                    "no_existing_file_expected": False,
                },
                "key_file": {
                    "content": "key file content",
                    "no_existing_file_expected": False,
                }
            }
            return "call result"

        mock_env = mock.MagicMock()
        booth_conf_middleware = env.middleware_config(
            "booth-name",
            "/local/file/path.conf",
            "/local/file/path.key",
        )

        self.assertEqual(
            "call result",
            booth_conf_middleware(next_in_line, mock_env)
        )

        self.assertEqual(self.read.mock_calls, [
            mock.call('/local/file/path.conf'),
            mock.call('/local/file/path.key', is_binary=True),
        ])

        self.assertEqual(self.process_no_existing_file_expectation.mock_calls, [
            mock.call(
                'booth config file',
                {
                    'content': 'file content',
                    'no_existing_file_expected': False
                },
                '/local/file/path.conf'
            ),
            mock.call(
                'booth key file',
                {
                    'content': 'key file content',
                    'no_existing_file_expected': False
                },
                '/local/file/path.key'
            ),
        ])

        self.assertEqual(self.write.mock_calls, [
            mock.call(
                {
                    'content': 'key file content',
                    'no_existing_file_expected': False
                },
                '/local/file/path.key'
            ),
            mock.call(
                {
                    'content': 'file content',
                    'no_existing_file_expected': False
                },
                '/local/file/path.conf'
            )
        ])

    def test_catch_exactly_his_exception(self):
        report_missing = self.setup_patch("env_file.report_missing")
        next_in_line = mock.Mock(side_effect=LibraryEnvError(
            ReportItem.error(report_codes.FILE_DOES_NOT_EXIST, info={
                "file_role": env_file_role_codes.BOOTH_CONFIG,
            }),
            ReportItem.error(report_codes.FILE_DOES_NOT_EXIST, info={
                "file_role": env_file_role_codes.BOOTH_KEY,
            }),
            ReportItem.error("OTHER ERROR", info={}),
        ))
        mock_env = mock.MagicMock()
        self.read.return_value = {"content": None}

        booth_conf_middleware = env.middleware_config(
            "booth-name",
            "/local/file/path.conf",
            "/local/file/path.key",
        )
        raised_exception = []
        def run_middleware():
            try:
                booth_conf_middleware(next_in_line, mock_env)
            except Exception as e:
                raised_exception.append(e)
                raise e
        self.assertRaises(LibraryEnvError, run_middleware)
        self.assertEqual(1, len(raised_exception[0].unprocessed))
        self.assertEqual("OTHER ERROR", raised_exception[0].unprocessed[0].code)
        self.assertEqual(report_missing.mock_calls, [
            mock.call('Booth config file', '/local/file/path.conf'),
            mock.call('Booth key file', '/local/file/path.key'),
        ])



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from lxml import etree

from pcs.lib import reports
from pcs.lib.cib.nvpair import update_nvset
from pcs.lib.cib.tools import get_nodes, find_unique_id
from pcs.lib.errors import LibraryError


def update_node_instance_attrs(cib, node_name, attrs, state_nodes=None):
    """
    Update nvpairs in instance_attributes for a node specified by its name.

    Automatically creates instance_attributes element if needed. If the node has
    more than one instance_attributes element, the first one is modified. If the
    node is missing in the CIB, it is automatically created if its state is
    provided in state_nodes.

    etree cib -- cib
    string node_name -- name of the node to be updated
    dict attrs -- attrs to update, e.g. {'A': 'a', 'B': ''}
    iterable state_nodes -- optional list of node state objects
    """
    node_el = _ensure_node_exists(get_nodes(cib), node_name, state_nodes)
    # If no instance_attributes id is specified, crm_attribute modifies the
    # first one found. So we just mimic this behavior here.
    attrs_el = node_el.find("./instance_attributes")
    if attrs_el is None:
        attrs_el = etree.SubElement(
            node_el,
            "instance_attributes",
            id=find_unique_id(cib, "nodes-{0}".format(node_el.get("id")))
        )
    update_nvset(attrs_el, attrs)

def _ensure_node_exists(tree, node_name, state_nodes=None):
    """
    Make sure node with specified name exists in the tree.

    If the node doesn't exist, raise LibraryError. If state_nodes is provided
    and contains state of a node with the specified name, create the node in
    the tree. Return existing or created node element.

    etree tree -- node parent element
    string name -- node name
    iterable state_nodes -- optional list of node state objects
    """
    node_el = _get_node_by_uname(tree, node_name)
    if node_el is None and state_nodes:
        for node_state in state_nodes:
            if node_state.attrs.name == node_name:
                node_el = _create_node(
                    tree,
                    node_state.attrs.id,
                    node_state.attrs.name,
                    node_state.attrs.type
                )
                break
    if node_el is None:
        raise LibraryError(reports.node_not_found(node_name))
    return node_el

def _get_node_by_uname(tree, uname):
    """
    Return a node element with specified uname in the tree or None if not found

    etree tree -- node parent element
    string uname -- node name
    """
    return tree.find("./node[@uname='{0}']".format(uname))

def _create_node(tree, node_id, uname, node_type=None):
    """
    Create new node element as a direct child of the tree element

    etree tree -- node parent element
    string node_id -- node id
    string uname -- node name
    string node_type -- optional node type (normal, member, ping, remote)
    """
    node = etree.SubElement(tree, "node", id=node_id, uname=uname)
    if node_type:
        node.set("type", node_type)
    return node




from __future__ import (
    absolute_import,
    division,
    print_function,
)

from pcs.test.tools.pcs_unittest import TestCase, skip


class BoothSendConfig(TestCase):
    """
    tested in:
        pcs.lib.commands.test.test_booth.ConfigSyncTest
    """

class BoothGetConfig(TestCase):
    """
    tested in:
        pcs.lib.commands.test.test_booth.PullConfigSuccess
        pcs.lib.commands.test.test_booth.PullConfigFailure
        pcs.lib.commands.test.test_booth.PullConfigWithAuthfileSuccess
        pcs.lib.commands.test.test_booth.PullConfigWithAuthfileFailure
    """


@skip("TODO: missing tests for pcs.lib.communication.booth.BoothSaveFiles")
class BoothSaveFiles(TestCase):
    def test_skip(self):
        pass
